from nerwhal.recognizer_bases.re_recognizer import ReRecognizer


class NumberRecognizer(ReRecognizer):
    """Recognize anything that is a number.

    This recognizer searches for digits, that might be separated by commas or dots. Having a low score, it can be used as a
    fallback to catch any number that hasn't been recognized by a more specific recognizer.
    """

    TAG = "NUMBER"
    SCORE = 0.5  # give it a low score so that it doesn't overwrite other entities

    @property
    def regexp(self):
        return r"\d[,\.\d]*\d?"



from nerwhal.backends.entity_ruler_backend import EntityRulerBackend
from nerwhal.recognizer_bases import EntityRulerRecognizer


def test_single_recognizer(embed):
    class TestRecognizer(EntityRulerRecognizer):
        TAG = "TEST"
        SCORE = 1.0

        @property
        def patterns(self):
            return [[{"TEXT": "abc"}, {"TEXT": "cde"}]]

    backend = EntityRulerBackend("de")
    backend.register_recognizer(TestRecognizer)
    text = "Das ist abc cde."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist TEST."
    assert ents[0].start_char == 8
    assert ents[0].end_char == 15
    assert ents[0].tag == "TEST"
    assert ents[0].text == "abc cde"
    assert ents[0].score == 1.0
    assert ents[0].recognizer == "TestRecognizer"


def test_multiple_recognizers(embed):
    class TestRecognizerA(EntityRulerRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def patterns(self):
            return [[{"TEXT": "abc"}]]

    class TestRecognizerB(EntityRulerRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def patterns(self):
            return [[{"TEXT": "cde"}]]

    backend = EntityRulerBackend("de")
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    text = "Das ist abc und cde."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist A und B."
    assert ents[0].tag == "A"
    assert ents[0].score == 1.0
    assert ents[1].tag == "B"
    assert ents[1].score == 0.5


def test_overlapping_recognizers(embed):
    class TestRecognizerA(EntityRulerRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def patterns(self):
            return [[{"TEXT": "abc"}, {"TEXT": "cde"}]]

    class TestRecognizerB(EntityRulerRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def patterns(self):
            return [[{"TEXT": "cde"}, {"TEXT": "efg"}]]

    class TestRecognizerC(EntityRulerRecognizer):
        TAG = "C"
        SCORE = 0.7

        @property
        def patterns(self):
            return [[{"TEXT": "cde"}, {"TEXT": "efg"}, {"TEXT": "ghi"}]]

    backend = EntityRulerBackend("de")
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    backend.register_recognizer(TestRecognizerC)
    text = "Das sind abc cde efg und abc cde efg ghi."
    ents = backend.run(text)
    #  The first match wins
    assert embed(text, ents) == "Das sind A efg und A efg ghi."



import importlib.util
import os
from collections import OrderedDict
from pathlib import Path
from typing import List

import nerwhal.backends
from nerwhal.combination_strategies import combine
from nerwhal.backends.stanza_ner_backend import StanzaNerBackend
from nerwhal.tokenizer import Tokenizer
from nerwhal.scorer import score_entities
from nerwhal.types import Config, NamedEntity
from nerwhal.entity_aligner import EntityAligner


class Analyzer:
    """The Analyzer class is NERwhal's core component, that runs named-entity recognition and caches the recognizers."""

    def __init__(self):
        self.config = None
        self.backends = OrderedDict()
        self.tokenizer = None
        self.recognizer_lookup = None

    def update_config(self, config):
        """Whenever one field of the config changed its value, the Analyzer's state is rebuilt from scratch.

        While reloading every recognizer and backend from scratch isn't the most efficient way, it is simple.
        """
        if config == self.config:
            return

        self.config = config

        self.tokenizer = Tokenizer(self.config.language)

        self.backends = {}

        if self.config.use_statistical_ner:
            self.backends["stanza"] = StanzaNerBackend(self.config.language)

        if self.config.load_integrated_recognizers:
            self._add_integrated_recognizers_to_config_recognizer_paths()

        self.recognizer_lookup = {}
        for recognizer_path in self.config.recognizer_paths:
            if not os.path.isfile(recognizer_path):
                raise ValueError(f"Configured recognizer {recognizer_path} is not a file")

            recognizer_cls = self._load_class(recognizer_path)
            self.recognizer_lookup[recognizer_cls.__name__] = recognizer_cls

            # import only the backend modules that are configured
            if recognizer_cls.BACKEND not in self.backends.keys():
                backend_cls = nerwhal.backends.load(recognizer_cls.BACKEND)

                if recognizer_cls.BACKEND == "entity-ruler":
                    backend_inst = backend_cls(self.config.language)
                else:
                    backend_inst = backend_cls()

                self.backends[recognizer_cls.BACKEND] = backend_inst

            self.backends[recognizer_cls.BACKEND].register_recognizer(recognizer_cls)

    def run_recognition(self, text):
        """Run recognition and return the recognized named-entities."""
        return self._run_backends(self.backends.values(), text)

    def _load_class(self, recognizer_path):
        """Load the class of a (custom) recognizer.

        This class may be located outside of NERwhal's package.
        """
        module_name = os.path.splitext(os.path.basename(recognizer_path))[0]
        spec = importlib.util.spec_from_file_location(module_name, recognizer_path)
        module = importlib.util.module_from_spec(spec)
        class_name = "".join(word.title() for word in module_name.split("_"))
        spec.loader.exec_module(module)
        recognizer_cls = getattr(module, class_name)
        return recognizer_cls

    def _add_integrated_recognizers_to_config_recognizer_paths(self):
        """Traverse the directory of integrated recognizers and append them to the recognizer paths."""
        for root, _, files in os.walk(Path(__file__).parent / "integrated_recognizers"):
            for file in files:
                if file.endswith("_recognizer.py"):
                    example = os.path.join(root, file)
                    if all([example not in path for path in self.config.recognizer_paths]):
                        self.config.recognizer_paths.append(example)

    def _run_backends(self, backends, text):
        return [backend.run(text) for backend in backends]


analyzer = Analyzer()


def recognize(text: str, config: Config, combination_strategy=None, context_words=False, return_tokens=True) -> dict:
    """Find personally identifiable data in the given text and return it.

    :param text: the text that is searched for named entities
    :param config: pass a config object to configure the recognition methods
    :param combination_strategy: choose from None, `disjunctive_union`, `fusion`, and `smart-fusion`; see the docs of
        `combination_strategies.combine` for more details
    :param context_words: if True, use context words to boost the score of entities: this is the case, if one of a recognizer's
        context words appears in the entity's sentence. Setting `context_words` to True will also align each entity's
        start/end to the nearest token's start/end
    :param return_tokens: compute and return the tokenization; this will also align each entity's start/end to the nearest
        token's start/end
    """
    analyzer.update_config(config)
    recognition_results = analyzer.run_recognition(text)

    if len(recognition_results) == 0:
        ents = []
    else:
        ents = [ent for result in recognition_results for ent in result]

    result = {}
    tokens = []
    if return_tokens or context_words:
        # tokenize
        analyzer.tokenizer.tokenize(text)
        tokens = analyzer.tokenizer.get_tokens()
        # align entities with tokens
        entity_aligner = EntityAligner()
        entity_aligner.align_entities_with_tokens(ents, tokens)

    # combine entities after they have been aligned
    ents = combine(ents, strategy=combination_strategy)

    if return_tokens:
        result["tokens"] = tokens

    if context_words:
        for ent in ents:
            sentence_tokens = analyzer.tokenizer.get_sentence_for_token(
                ent.start_tok, exclude_tokens=list(range(ent.start_tok, ent.end_tok))
            )
            sentence_words = [token.text for token in sentence_tokens]
            context_words = analyzer.recognizer_lookup[ent.recognizer].CONTEXT_WORDS
            if any(word in sentence_words for word in context_words):
                ent.score = min(ent.score * analyzer.config.context_word_confidence_boost_factor, 1.0)

    result["ents"] = ents
    return result


def evaluate(ents: List[NamedEntity], gold: List[NamedEntity]) -> dict:
    """Compute the evaluation scores of computed named entities compared to the underlying true entities.

    Each entity has to have the fields `start_char`, `end_char` and `tag` populated. The remaining fields are ignored.
    """
    return score_entities(ents, gold)



from dataclasses import dataclass, field
from typing import List, Optional
import pydantic


@dataclass
class NamedEntity:
    start_char: int
    end_char: int
    tag: str
    text: str = None
    score: float = None
    recognizer: str = None
    start_tok: int = None
    end_tok: int = None


@dataclass
class Token:
    """
    Fields:
        has_ws: whether or not the token is followed by whitespace (useful when reconstructing the text from tokens)
        br_count: some tokens represented breaks/newlines (e.g. "\n\n"), break count specifies how many breaks there are;
            defaults to 0 for all other tokens
    """

    text: str
    has_ws: bool
    br_count: int
    start_char: int
    end_char: int


@pydantic.dataclasses.dataclass
class Config:
    """The main config object.

    Fields:
        language: the language of the text; used for tokenization and statistical NER
        recognizer_paths: a list of paths that specify the (custom) recognizers to be used for recognition (optional)
        use_statistical_ner: whether to use statistical NER or not (optional)
        load_integrated_recognizers: whether to use all integrated recognizers for recognition (optional)
        context_word_confidence_boost_factor: the factor by which the score of named entities gets increased, if one of the
            context words is nearby (optional)
    """

    language: str
    recognizer_paths: Optional[List[str]] = field(default_factory=list)
    use_statistical_ner: Optional[bool] = False
    load_integrated_recognizers: Optional[bool] = False
    context_word_confidence_boost_factor: Optional[float] = 1.2



class EntityAligner:
    """This class aligns the start and end characters of entities with a given tokenization.

    In general, entity recognition methods do not know about token borders and thus do not always align with the tokenization.
    The purpose of this class is to align the entities with the nearest larger token border, and to set the entities start and
    end token indices.
    """

    def __init__(self):
        self._char_to_token_idx = {}

    def align_entities_with_tokens(self, ents, tokens):
        """Alter the given entity objects to match the nearest larger token start and end char, and set their start and end
        token indices.

        :param ents: a list of NamedEntity objects
        :param tokens: a list of token objects
        """
        self._compute_char_to_token_idx_lookup(tokens)
        self._set_tokens_in_entities(ents)

    def _set_tokens_in_entities(self, ents):
        """Set the start and end token indices in the given entities.

        Note, that this may also alter the entities start and end characters.
        """
        for ent in ents:
            self._set_start_tok(ent)
            self._set_end_tok(ent)

    def _set_start_tok(self, ent):
        """Recursive function to align the start of the entity with the nearest token border."""
        try:
            ent.start_tok = self._char_to_token_idx[ent.start_char]
        except KeyError:
            # search for the nearest larger token border
            ent.start_char -= 1
            self._set_start_tok(ent)

    def _set_end_tok(self, ent):
        """Recursive function to align the end of the entity with the nearest token border."""
        try:
            # take the last character in the entity to get the token index and to create a slice take the next token
            ent.end_tok = self._char_to_token_idx[ent.end_char - 1] + 1
        except KeyError:
            # search for the nearest larger token border
            ent.end_char += 1
            self._set_end_tok(ent)

    def _compute_char_to_token_idx_lookup(self, tokens):
        """Computes a lookup table, that maps the first and last character of each token to its index in the document."""
        self._char_to_token_idx = {}

        char = 0
        for idx, token in enumerate(tokens):
            token_len = len(token.text)
            self._char_to_token_idx[char] = idx  # first character in token
            self._char_to_token_idx[char + token_len - 1] = idx  # last character in token

            char += token_len
            if token.has_ws:
                char += 1



from spacy.scorer import PRFScore


def score_entities(candidates, gold):
    """Compute a range of scores for a list of named entities compared to the true entities contained in the text.

    :param candidates: a list of named entities, with at least have the properties `start_char`, `end_char`, and `tag`
    :param gold: the true list of named entities, with at least have the properties `start_char`, `end_char`, and `tag`
    :return: a dictionary with the scores
    """
    cand_tuples = _to_start_end_tag_tuples(candidates)
    gold_tuples = _to_start_end_tag_tuples(gold)
    tags = set([tag for _, _, tag in gold_tuples + cand_tuples])

    scores = {"total": score(cand_tuples, gold_tuples), "tags": {}}
    for tag in tags:
        scores["tags"][tag] = score(_select_tuples_with_tag(tag, cand_tuples), _select_tuples_with_tag(tag, gold_tuples))

    return scores


def score(candidates, gold):
    """Compute the most common scoring measures for any two sets.

    :param candidates: the candidates to be scored, i.e. the predicted values
    :param gold: the true values
    :return: a dictionary with the scores
    """

    prf_scorer = PRFScore()
    prf_scorer.score_set(set(candidates), set(gold))

    return {
        "f1": prf_scorer.fscore,
        "f2": _fbeta_score(2, prf_scorer.precision, prf_scorer.recall),
        "precision": prf_scorer.precision,
        "recall": prf_scorer.recall,
        "true_positives": prf_scorer.tp,
        "false_positives": prf_scorer.fp,
        "false_negatives": prf_scorer.fn,
    }


def _to_start_end_tag_tuples(ents):
    tuples = [(ent.start_char, ent.end_char, ent.tag) for ent in ents]
    return tuples


def _select_tuples_with_tag(tag, start_end_tag_tuples):
    return [_tuple for _tuple in start_end_tag_tuples if _tuple[2] == tag]


def _fbeta_score(beta, p, r):
    """Compute the F-beta score for a given precision and recall."""
    return (1 + beta ** 2) * (p * r) / ((beta ** 2 * p) + r + 1e-100)



import pytest

from nerwhal.integrated_recognizers.number_recognizer import NumberRecognizer


@pytest.fixture
def backend(setup_backend):
    recognizer = NumberRecognizer
    backend = setup_backend(recognizer.BACKEND)
    backend.register_recognizer(recognizer)
    return backend


def test_number(backend, embed):
    text = "Chewbacca wiegt 102 kg."
    ents = backend.run(text)
    assert embed(text, ents) == "Chewbacca wiegt NUMBER kg."


def test_number_wo_space(backend, embed):
    text = "Er fliegt mit 10000km/h."
    ents = backend.run(text)
    assert embed(text, ents) == "Er fliegt mit NUMBERkm/h."


def test_number_w_dot(backend, embed):
    text = "Er fliegt mit 1.000.000 km/h."
    ents = backend.run(text)
    assert embed(text, ents) == "Er fliegt mit NUMBER km/h."


def test_number_w_comma(backend, embed):
    text = "Das maximale Gewicht für Handgepäck ist 9,99 kg."
    ents = backend.run(text)
    assert embed(text, ents) == "Das maximale Gewicht für Handgepäck ist NUMBER kg."



from nerwhal.recognizer_bases.entity_ruler_recognizer import EntityRulerRecognizer


class MoneyRecognizer(EntityRulerRecognizer):
    """Recognize money amounts written as a number followed by a currency symbol.

    Attention: This recognizer does only match one of the manifold ways that money amounts can be written.

    For example, currency codes or currency names are not recognized.
    """

    TAG = "MONEY"
    SCORE = 0.99

    @property
    def patterns(self):
        return [
            [{"TEXT": {"REGEX": r"^\d[,\.\d]+\d$"}}, {"IS_CURRENCY": True}],
            [{"IS_CURRENCY": True}, {"TEXT": {"REGEX": r"^\d[,\.\d]+\d$"}}],
        ]



import subprocess


def load_stanza_nlp(language, processors):
    """Load the Stanza nlp object.

    If the language's models cannot be found, they are downloaded.
    """
    import stanza

    try:
        stanza_nlp = stanza.Pipeline(lang=language, processors=processors)
    except (Exception, SystemExit):
        stanza.download(language, processors=processors)
        stanza_nlp = stanza.Pipeline(lang=language, processors=processors)

    return stanza_nlp


def load_spacy_nlp(language, disable_components):
    """Load the spaCy nlp object.

    If the language's models cannot be found, they are downloaded.
    """
    import spacy

    try:
        spacy_nlp = spacy.load(language, disable=disable_components)
    except OSError:
        subprocess.run(["python", "-m", "spacy", "download", language])
        spacy_nlp = spacy.load(language, disable=disable_components)

    return spacy_nlp


def configure_spacy_entity_extension_attributes():
    """Add custom extension attributes to the spaCy Span class."""
    from spacy.tokens import Span

    if not Span.has_extension("score"):
        Span.set_extension("score", default=-1.0)
    if not Span.has_extension("recognizer"):
        Span.set_extension("recognizer", default="")


def set_spacy_entity_extension_attributes(score, recognizer):
    """Returns a spaCy pipeline component, that sets the score and recognizer in a doc's entities."""

    def function(doc):
        for ent in doc.ents:
            if ent._.score >= 0:
                continue
            ent._.score = score
            ent._.recognizer = recognizer
        return doc

    return function



from nerwhal.core import recognize, evaluate  # noqa: F401
from nerwhal.types import NamedEntity, Config  # noqa: F401
import nerwhal.recognizer_bases  # noqa: F401
from nerwhal.integrated_recognizers import list_all as list_integrated_recognizers  # noqa: F401



from .base import Backend
from nerwhal.types import NamedEntity
from nerwhal.nlp_utils import load_stanza_nlp

# the stanza NER models have an F1 score between 74.3 and 94.8, https://stanfordnlp.github.io/stanza/performance.html
# we choose a hardcoded score in this scale
NER_SCORE = 0.8


class StanzaNerBackend(Backend):
    """This backend recognizes entities using Stanza's neural network models.

    See https://stanfordnlp.github.io/stanza/ for more information on Stanza.

    Work in process:
    - this backend does not yet support different recognizers (i.e. custom models)
    - instead it uses the default Stanza model for the provided language
    """

    def __init__(self, language):
        self.stanza_nlp = load_stanza_nlp(language, processors="tokenize,mwt,ner")

    def register_recognizer(self, recognizer_cls):
        raise NotImplementedError()

    def run(self, text):
        doc = self.stanza_nlp(text)
        return [
            NamedEntity(ent.start_char, ent.end_char, ent.type, ent.text, NER_SCORE, self.__class__.__name__)
            for ent in doc.entities
        ]



from nerwhal.recognizer_bases.entity_ruler_recognizer import EntityRulerRecognizer


class DeDateRecognizer(EntityRulerRecognizer):
    """Recognize dates (without time) in common German formats.

    The recognizer aims at matching date formats in the DIN 1355-1 and DIN 5008 formats, as well as in the commonly used
    "DD. Monat YYYY" format. For details on what is matched, please have a look at the pattern definition.
    """

    TAG = "DATE"
    SCORE = 0.99
    CONTEXT_WORDS = ["Datum", "Tag", "Geburtstag", "Geburtsdatum"]

    @property
    def patterns(self):
        return [
            [  # DIN 1355-1, e.g. 24.12.2020
                {"TEXT": {"REGEX": r"^(0?[1-9]|[1-2][0-9]|3[01])\.(0?[1-9]|1[0-2])\.[1-2]?[0-9]?[0-9][0-9]$"}},
            ],
            [  # DIN 5008, e.g. 2020-12-24
                {"TEXT": {"REGEX": r"^[1-2][0-9]{3}$"}},
                {"TEXT": "-"},
                {"TEXT": {"REGEX": r"^(0[1-9]|1[0-2])$"}},
                {"TEXT": "-"},
                {"TEXT": {"REGEX": r"^(0[1-9]|[1-2][0-9]|3[01])$"}},
            ],  # Written out, e.g. 24. Aug. 2020, optional with day Mittwoch, 24. Aug. 2020
            [
                {"TEXT": {"REGEX": r"^Montag|Dienstag|Mittwoch|Donnerstag|Freitag|Samstag|Sonntag"}, "OP": "?"},
                {"IS_ALPHA": False, "OP": "?"},
                {"TEXT": {"REGEX": r"^(0?[1-9]|[1-2][0-9]|3[01])\.$"}},
                {"TEXT": {"REGEX": r"^Jan|Feb|Mrz|Mä|Apr|Mai|Jun|Jul|Aug|Sep|Okt|Nov|Dez"}},
                {"TEXT": {"REGEX": r"^([1-2][0-9]{3}|[1-9][0-9]{2}|[0-9]{2})$"}},
            ],
        ]



from nerwhal.backends.flashtext_backend import FlashtextBackend
from nerwhal.recognizer_bases import FlashtextRecognizer


def test_single_recognizer(embed):
    class TestRecognizer(FlashtextRecognizer):
        TAG = "XX"
        SCORE = 1.0

        @property
        def keywords(self):
            return ["abc", "cde"]

    backend = FlashtextBackend()
    backend.register_recognizer(TestRecognizer)
    text = "Das ist abc und cde."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist XX und XX."
    assert ents[0].start_char == 8
    assert ents[0].end_char == 11
    assert ents[0].tag == "XX"
    assert ents[0].text == "abc"
    assert ents[0].score == 1.0
    assert ents[0].recognizer == "TestRecognizer"


def test_multiple_recognizers(embed):
    class TestRecognizerA(FlashtextRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def keywords(self):
            return ["abc"]

    class TestRecognizerB(FlashtextRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def keywords(self):
            return ["cde"]

    backend = FlashtextBackend()
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    text = "Das ist abc und cde."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist A und B."
    assert ents[0].tag == "A"
    assert ents[0].score == 1.0
    assert ents[1].tag == "B"
    assert ents[1].score == 0.5


def test_overlapping_recognizers(embed):
    class TestRecognizerA(FlashtextRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def keywords(self):
            return ["abc", "cde"]

    class TestRecognizerB(FlashtextRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def keywords(self):
            return ["cde", "fgh"]

    backend = FlashtextBackend()
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    text = "Das ist cde."
    ents = backend.run(text)
    #  Recognizer B overwrites the keyword "cde"
    assert embed(text, ents) == "Das ist B."



from nerwhal.backends.re_backend import ReBackend
from nerwhal.recognizer_bases import ReRecognizer


def test_single_recognizer(embed):
    class TestRecognizer(ReRecognizer):
        TAG = "TEST"
        SCORE = 1.0

        @property
        def regexp(self):
            return r"abc"

    backend = ReBackend()
    backend.register_recognizer(TestRecognizer)
    text = "Das ist abc und abc."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist TEST und TEST."
    assert ents[0].start_char == 8
    assert ents[0].end_char == 11
    assert ents[0].tag == "TEST"
    assert ents[0].text == "abc"
    assert ents[0].score == 1.0
    assert ents[0].recognizer == "TestRecognizer"


def test_multiple_recognizers(embed):
    class TestRecognizerA(ReRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def regexp(self):
            return r"abc"

    class TestRecognizerB(ReRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def regexp(self):
            return r"123"

    backend = ReBackend()
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    text = "Das ist abc und 123."
    ents = backend.run(text)
    assert embed(text, ents) == "Das ist A und B."
    assert ents[0].tag == "A"
    assert ents[0].score == 1.0
    assert ents[1].tag == "B"
    assert ents[1].score == 0.5


def test_overlapping_recognizers(embed):
    class TestRecognizerA(ReRecognizer):
        TAG = "A"
        SCORE = 1.0

        @property
        def regexp(self):
            return r"abc"

    class TestRecognizerB(ReRecognizer):
        TAG = "B"
        SCORE = 0.5

        @property
        def regexp(self):
            return r"ab"

    backend = ReBackend()
    backend.register_recognizer(TestRecognizerA)
    backend.register_recognizer(TestRecognizerB)
    text = "Das ist abc."
    ents = backend.run(text)
    assert ents[0].start_char == 8
    assert ents[0].end_char == 11
    assert ents[0].tag == "A"
    assert ents[0].text == "abc"

    assert ents[1].start_char == 8
    assert ents[1].end_char == 10
    assert ents[1].tag == "B"
    assert ents[1].text == "ab"



import csv
from pathlib import Path

from nerwhal.recognizer_bases.flashtext_recognizer import FlashtextRecognizer


class DeCountryRecognizer(FlashtextRecognizer):
    """Recognize German country names in short and long form.

    This recognizer searches for occurrences of German country names, loaded from a file. Thus, what it can find is limited by
    the completeness of the data in this file. Our data file also contains many declined versions of the long form of country
    names, that can then be recognized as well.
    """

    TAG = "COUNTRY"
    SCORE = 0.95
    CONTEXT_WORDS = ["Land", "Länder", "Staat", "Staaten"]

    def __init__(self):
        path = Path(__file__).parent / "data" / "countries.csv"
        with open(path) as csv_file:
            reader = csv.reader(csv_file, delimiter=";")
            self.country_names = [name for row in reader for name in row]

    @property
    def keywords(self):
        return self.country_names



import re
from typing import Type

from .base import Backend
from nerwhal.types import NamedEntity
from ..recognizer_bases import ReRecognizer


class ReBackend(Backend):
    """This backend recognizes entities using Python's regular expression engine.

    See https://docs.python.org/3.7/library/re.html for more information.
    """

    def __init__(self):
        self.compiled_regexps = []
        self.recognizer_classes = []

    def register_recognizer(self, recognizer_cls: Type[ReRecognizer]):
        recognizer = recognizer_cls()

        self.compiled_regexps += [re.compile(recognizer.regexp, flags=recognizer.FLAGS)]
        self.recognizer_classes.append(recognizer_cls)

    def run(self, text):
        ents = []
        for pattern, rec in zip(self.compiled_regexps, self.recognizer_classes):
            ents += [
                NamedEntity(*m.span(rec.GROUP), rec.TAG, m.group(rec.GROUP), rec.SCORE, rec.__name__)
                for m in pattern.finditer(text)
            ]
        return ents



from distutils.core import setup

from setuptools import find_packages

setup(
    name="nerwhal",
    version="0.1.0",
    url="https://openredact.org/",
    author="Jonas Langhabel",
    author_email="hello@openredact.org",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    license="MIT",
    description="Find personally identifiable information in German texts using NER and rule based matching.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "spacy>=2.3.2,<3.0.0",
        "stanza>=1.0.1,<1.1.0",
        "torch>=1.3.0,<1.6.0",
        "flashtext>=2.7,<3.0",
        "pydantic>=1.0.0,<2.0.0",
    ],
    python_requires=">=3.7",
)



import pytest

from nerwhal.integrated_recognizers.de.de_country_recognizer import DeCountryRecognizer


@pytest.fixture(scope="module")
def backend(setup_backend):
    recognizer = DeCountryRecognizer
    backend = setup_backend(recognizer.BACKEND)
    backend.register_recognizer(recognizer)
    return backend


def test_single_word_country(backend, embed):
    text = "Finnland hat ein neues Gesetz erlassen."
    ents = backend.run(text)
    assert embed(text, ents) == "COUNTRY hat ein neues Gesetz erlassen."


def test_multi_word_country(backend, embed):
    text = "Was ist die Hauptstadt der Republik Nordmazedonien?"
    ents = backend.run(text)
    assert embed(text, ents) == "Was ist die Hauptstadt der COUNTRY?"


def test_declined_multi_word_country(backend, embed):
    text = "Was ist die Hauptstadt der Demokratischen Volksrepublik Laos?"
    ents = backend.run(text)
    assert embed(text, ents) == "Was ist die Hauptstadt der COUNTRY?"



import pytest


@pytest.fixture(scope="session")
def setup_backend():
    def function(backend: str, **kwargs):
        """Create a backend for the given recognizer and register the recognizer at it."""
        if backend == "re":
            from nerwhal.backends.re_backend import ReBackend

            backend = ReBackend()
        elif backend == "flashtext":
            from nerwhal.backends.flashtext_backend import FlashtextBackend

            backend = FlashtextBackend()
        elif backend == "entity-ruler":
            from nerwhal.backends.entity_ruler_backend import EntityRulerBackend

            backend = EntityRulerBackend(**kwargs)
        else:
            raise ValueError(f"Unknown backend {backend}")
        return backend

    return function



from abc import ABC, abstractmethod
from typing import Type

from nerwhal.recognizer_bases.base import Recognizer


class Backend(ABC):
    """This is the base class for all recognizer backends.

    Backends are the engines behind the recognizers that do the actual hard lifting, i.e. they run the search for named
    entities. Each backend can run several recognizers.
    """

    @abstractmethod
    def register_recognizer(self, recognizer_cls: Type[Recognizer]):
        """Add the given recognizer to this backend instance.

        One backend can have several recognizers. Once a recognizer is added, it cannot be removed (create a new backend
        instance if you have to remove it).
        """
        pass

    @abstractmethod
    def run(self, text):
        """Run the backend using all registered recognizers.

        :return: the list of named entities that have been identified
        """
        pass
